# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from __future__ import absolute_import

from .errors import InvalidChecksumError
from .errors import StreamingError
from .messages import ErrorCode
from .messages import PingResponseMessage
from .messages import Types


class RequestHandler(object):
    """Specifies how a TChannel server handles incoming requests.

    This class is agnostic of whether specific implementations are synchronous
    or Tornado-based.
    """

    def handle(self, message, connection):
        """Handle an incoming request.

        The handshake has already been completed.

        :param message:
            the incoming message.
        :param connection:
            Reference to the connection object
        :returns:
            Nothing. The connection object must be used to send the response
            back.
        """
        raise NotImplementedError()


class BaseRequestHandler(RequestHandler):
    """A minimal RequestHandler skeleton.

    This implements a minimal base RequestHandler that implements methods
    which should be mostly the same between implementations. Implementations
    must inherit this class and override at least ``handle_call``.
    """

    _HANDLER_NAMES = {
        Types.PING_REQ: 'ping',
        Types.CALL_REQ: 'pre_call',
        Types.CALL_REQ_CONTINUE: 'pre_call'
    }

    def __init__(self):
        super(BaseRequestHandler, self).__init__()

    def handle(self, message, connection):
        # TODO assert that the handshake was already completed
        assert message, "message must not be None"

        if message.message_type not in self._HANDLER_NAMES:
            # TODO handle this more gracefully
            raise NotImplementedError("Unexpected message: %s" % str(message))

        handler_name = "handle_" + self._HANDLER_NAMES[message.message_type]
        return getattr(self, handler_name)(message, connection)

    def handle_pre_call(self, message, connection):
        """Handle incoming request message including CallRequestMessage and
        CallRequestContinueMessage

        This method will build the User friendly request object based on the
        incoming messages.

        It passes all the messages into the message_factory to build the init
        request object. Only when it get a CallRequestMessage and a completed
        arg_1=argstream[0], the message_factory will return a request object.
        Then it will trigger the async call_handle call.

        :param message: CallRequestMessage or CallRequestContinueMessage
        :param connection: tornado connection
        """
        try:
            req = connection.request_message_factory.build(message)
            # message_factory will create Request only when it receives
            # CallRequestMessage. It will return None, if it receives
            # CallRequestContinueMessage.
            if req:
                self.handle_call(req, connection)

        except (InvalidChecksumError, StreamingError) as e:
            connection.send_error(
                ErrorCode.bad_request,
                e.message,
                message.id,
            )
        except Exception:
            connection.send_error(
                ErrorCode.unexpected,
                "An unexpected error has occurred.",
                message.id,
            )

    def handle_ping(self, ping, connection):
        return connection.write(PingResponseMessage(), ping.id)

    def handle_call(self, call, connection):
        """Handle an incoming call.

        :param call:
            CallRequestMessage containing information about the call
        :param connection:
            Connection through which the call was made
        :returns:
            Nothing. The response must be sent using the
            implementation-specific connection object.
        """
        raise NotImplementedError("Must be implemented.")


class CallableRequestHandler(RequestHandler):
    """An adapter from a function to a RequestHandler."""

    def __init__(self, f):
        assert f
        self._f = f

    def handle(self, message, connection):
        return self._f(message, connection)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import inspect
import logging
import os
import sys
from functools import partial

import tornado.gen
import tornado.ioloop
import tornado.iostream
import tornado.tcpserver
from tornado.netutil import bind_sockets

from . import hyperbahn
from .. import scheme
from ..enum import enum
from ..event import EventEmitter
from ..event import EventRegistrar
from ..handler import CallableRequestHandler
from ..net import local_ip
from ..zipkin.zipkin_trace import ZipkinTraceHook
from .broker import ArgSchemeBroker
from .connection import StreamConnection
from .dispatch import RequestDispatcher
from .peer import PeerGroup

log = logging.getLogger('tchannel')


State = enum(
    'State',
    ready=0,
    closing=1,
    closed=2,
)


class TChannel(object):
    """Manages inbound and outbound connections to various hosts."""

    _SCHEMES = {
        'raw': scheme.RawArgScheme,
        'json': scheme.JsonArgScheme,
        # 'http': scheme.HttpArgScheme, TODO
    }

    def __init__(self, name, hostport=None, process_name=None,
                 known_peers=None, trace=False):
        """Build or re-use a TChannel.

        :param name:
            Name is used to identify client or service itself.
        :param hostport:
            The host-port at which the service behind this TChannel is
            reachable. The port specified in the ``hostport`` is what the
            server will listen on. If unspecified, the system will attempt to
            determine the local network IP for this host and use an
            OS-assigned port.
        :param process_name:
            Name of this process. This is used for logging only. If
            unspecified, this will default to ``$processName[$processId]``.
        :param known_peers:
            A list of host-ports at which already known peers can be reached.
            Defaults to an empty list.
        :param trace:
            Flag to turn on/off zipkin trace. It can be a bool variable or
            a function that return true or false.
        """
        self._state = State.ready
        self._handler = RequestDispatcher()

        self.peers = PeerGroup(self)

        self._port = 0
        self._host = None
        if hostport:
            self._host, port = hostport.rsplit(':', 1)
            self._port = int(port)

        if not self._host:
            # TChannel(":4040") => determine IP automatically but use port
            # 4040
            self._host = local_ip()

        self.process_name = process_name or "%s[%s]" % (
            sys.argv[0], os.getpid()
        )

        self.name = name
        self._trace = trace

        # register event hooks
        self.event_emitter = EventEmitter()
        self.hooks = EventRegistrar(self.event_emitter)
        self.hooks.register(ZipkinTraceHook(tchannel=self))

        if known_peers:
            for peer_hostport in known_peers:
                self.peers.add(peer_hostport)

    @property
    def trace(self):
        if callable(self._trace):
            return self._trace()
        else:
            return self._trace

    def advertise(self, router, name=None):
        """Advertise the given TChannel to Hyperbahn.

        This informs Hyperbahn that the given client/service is using TChannel
        at a fixed rate.

        It also tells the TChannel about the given Hyperbahn routers.

        :param routers:
            Seed list of addresses of Hyperbahn routers
        :param name:
            Name to be register on the hyperbahn.
        :returns:
            A future that resolves to the remote server's response after
            the first advertise finishes.
        """
        name = name or self.name
        return hyperbahn.advertise(self, name, router)

    @property
    def closed(self):
        return self._state == State.closed

    @tornado.gen.coroutine
    def close(self):
        if self._state in [State.closed, State.closing]:
            raise tornado.gen.Return(None)

        self._state = State.closing
        try:
            yield self.peers.clear()
        finally:
            self._state = State.closed

    @property
    def hostport(self):
        return "%s:%d" % (self._host, self._port)

    def request(self,
                hostport=None,
                service=None,
                arg_scheme=None,
                retry=None,
                **kwargs):
        """Initiate a new request through this TChannel.

        :param hostport:
            Host to which the request will be made. If unspecified,  a random
            known peer will be picked.
        :param service:
            Service being called. Defaults to an empty string.
        :param arg_scheme:
            Arg scheme type.
        :param rety:
            Retry flag
        """
        return self.peers.request(hostport=hostport,
                                  service=service,
                                  arg_scheme=arg_scheme,
                                  retry=retry,
                                  **kwargs)

    def listen(self, port=None):
        """Start listening for incoming connections.

        A request handler must have already been specified with
        ``TChannel.host``.

        :param port: the port that tchannel listens on
        """
        if port:
            assert not self._port, "Port has already been set."
            self._port = int(port)

        assert self._handler, "Call .host with a RequestHandler first"
        server = TChannelServer(self)

        sockets = bind_sockets(self._port)
        assert sockets, "No sockets bound for port %d" % self._port

        # If port was 0, the OS probably assigned something better.
        self._port = sockets[0].getsockname()[1]

        server.add_sockets(sockets)

    @tornado.gen.coroutine
    def receive_call(self, message, connection):
        if not self._handler:
            log.warn(
                "Received %s but a handler has not been defined.", str(message)
            )
            return
        self._handler.handle(message, connection)

    def _register_simple(self, endpoint, scheme, f):
        """Register a simple endpoint with this TChannel.

        :param endpoint:
            Name of the endpoint being registered.
        :param scheme:
            Name of the arg scheme under which the endpoint will be
            registered.
        :param f:
            Callable handler for the endpoint.
        """
        assert scheme in self._SCHEMES, ("Unsupported arg scheme %s" % scheme)
        scheme = self._SCHEMES[scheme]()
        self._handler.register(endpoint, f, ArgSchemeBroker(scheme))
        return f

    def _register_thrift(self, service_module, handler, **kwargs):
        """Register a Thrift endpoint on this TChannel.

        :param service_module:
            Reference to the Thrift-generated module for the service being
            registered.
        :param handler:
            Handler for the endpoint
        :param method:
            Name of the Thrift method being registered. If omitted, ``f``'s
            name is assumed to be the method name.
        :param service:
            Name of the Thrift service. By default this is determined
            automatically from the module name.
        """
        import tchannel.thrift as thrift
        # Imported inside the function so that we don't have a hard dependency
        # on the Thrift library. This function is usable only if the Thrift
        # library is installed.
        thrift.register(self._handler, service_module, handler, **kwargs)
        return handler

    def register(self, endpoint, scheme=None, handler=None, **kwargs):
        """Register a handler with this TChannel.

        This may be used as a function or as a decorator.

        .. code-block:: python

            app = TChannel()

            @app.register("hello", "json")
            def hello(request, response, tchannel):
                # ...

            app.register(Foo, "thrift", handler_func)

        :param endpoint:
            Name of the endpoint being registered for raw and JSON arg
            schemes. Reference to the Thrift-generated module for the Thrift
            arg scheme.
        :param scheme:
            Name of the scheme under which the endpoint is being registered.
            One of ``raw``, ``json``, and ``thrift``. Defaults to "raw",
            except if ``endpoint`` was a module, in which case this defaults
            to "thrift".
        :param handler:
            If specified, this is the handler function. If ignored, this
            function returns a decorator that can be used to register the
            handler function.
        :returns:
            If ``handler`` was specified, this returns ``handler``. Otherwise,
            it returns a decorator that can be applied to a function to
            register it as the handler.
        """
        assert endpoint, "endpoint is required"

        if not scheme:
            # scheme defaults to raw, unless the endpoint is a service module.
            if inspect.ismodule(endpoint):
                scheme = "thrift"
            else:
                scheme = "raw"
        scheme = scheme.lower()

        if scheme == 'thrift':
            decorator = partial(self._register_thrift, endpoint, **kwargs)
        else:
            decorator = partial(
                self._register_simple, endpoint, scheme, **kwargs
            )

        if handler is not None:
            return decorator(handler)
        else:
            return decorator


class TChannelServer(tornado.tcpserver.TCPServer):
    __slots__ = ('tchannel',)

    def __init__(self, tchannel):
        super(TChannelServer, self).__init__()
        self.tchannel = tchannel

    @tornado.gen.coroutine
    def handle_stream(self, stream, address):
        log.debug("New incoming connection from %s:%d" % address)

        conn = StreamConnection(connection=stream, tchannel=self.tchannel)

        yield conn.expect_handshake(headers={
            'host_port': self.tchannel.hostport,
            'process_name': self.tchannel.process_name,
        })

        log.debug(
            "Successfully completed handshake with %s:%s (%s)",
            conn.remote_host,
            conn.remote_host_port,
            conn.remote_process_name)

        self.tchannel.peers.get(
            "%s:%s" % (conn.remote_host,
                       conn.remote_host_port)
        ).register_incoming(conn)

        yield conn.serve(handler=CallableRequestHandler(self._handle))

    def _handle(self, message, connection):
        self.tchannel.receive_call(message, connection)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import


class TChannelError(Exception):
    """Represent a TChannel-generated exception."""
    pass


class ProtocolError(TChannelError):
    """Represent a protocol-level exception"""
    __slots__ = (
        'code',
        'description',
        'id',
        'tracing',
    )

    def __init__(
        self,
        code,
        description,
        id=None,
        tracing=None,
    ):
        super(TChannelError, self).__init__(description)
        self.code = code
        self.tracing = tracing
        self.id = id
        self.description = description


class InvalidMessageError(TChannelError):
    """Represent an invalid message."""
    pass


class InvalidEndpointError(TChannelError):
    """Represent an message containing invalid endpoint."""
    pass


class TimeoutError(TChannelError):
    # TODO fix circular dependence
    code = 0x01


class ConnectionClosedError(TChannelError):
    pass


class ReadError(TChannelError):
    """Raised when there is an error while reading input."""
    pass


class InvalidChecksumError(TChannelError):
    """Represent invalid checksum type in the message"""
    pass


class StreamingError(TChannelError):
    """Represent Streaming Message Exception"""
    pass


class NoAvailablePeerError(TChannelError):
    """Represents a failure to find any peers for a request."""
    pass


class InvalidErrorCodeError(TChannelError):
    """Represent Invalid Error Code exception"""
    def __init__(self, code):
        super(InvalidErrorCodeError, self).__init__(
            'Invalid Error Code (%s)' % (code))
        self.code = code


class AdvertiseError(TChannelError):
    """Represent advertise failure exception"""
    pass


class TChannelApplicationError(TChannelError):
    """The remote application returned an exception.

    This is not a protocol error. This means a response was received with the
    ``code`` flag set to fail.
    """
    def __init__(self, code, args):
        super(TChannelError, self).__init__(
            'TChannel application error (%s)' % (args)
        )

        self.code = code
        self.args = args



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import tornado
import tornado.gen

from ..enum import enum
from ..errors import TChannelError
from ..messages.common import FlagsType
from ..messages.common import StreamState
from .stream import InMemStream
from .util import get_arg

StatusCode = enum(
    'StatusCode',
    ok=0x00,
    error=0x01,
)


class Response(object):
    """An outgoing response.

    Response class is used to represent the CallResponseMessage at User's
    level. This is going to hide the protocol level message information.
    """

    # TODO decide which elements inside "message" object to expose to user.
    def __init__(
        self,
        connection=None,
        id=None,
        flags=None,
        code=None,
        tracing=None,
        headers=None,
        checksum=None,
        argstreams=None,
        scheme=None,
    ):

        self.flags = flags or FlagsType.none
        self.code = code or StatusCode.ok
        self.tracing = tracing
        self.checksum = checksum
        # argstreams is a list of InMemStream/PipeStream objects
        self.argstreams = argstreams or [InMemStream(),
                                         InMemStream(),
                                         InMemStream()]
        self.headers = headers or {}
        self.id = id
        self.connection = connection
        self.state = StreamState.init
        self.flushed = False

        self.scheme = scheme

    @property
    def status_code(self):
        return self.code

    @status_code.setter
    def status_code(self, status):
        if status not in StatusCode:
            raise TChannelError("Invalid status code!")

        self.code = status.value

    @property
    def ok(self):
        return self.code == StatusCode.ok.value

    def get_header_s(self):
        """Get the raw stream of header.

        :return: the argstream of header
        """
        return self.argstreams[1]

    def get_body_s(self):
        """Get the raw stream of body.

        :return: the argstream of body
        """
        return self.argstreams[2]

    @tornado.gen.coroutine
    def get_header(self):
        """Get the header value from the response.

        :return: a future contains the deserialized value of header
        """
        raw_header = yield get_arg(self, 1)
        if not self.scheme:
            raise tornado.gen.Return(raw_header)
        else:
            header = self.scheme.deserialize_header(raw_header)
            raise tornado.gen.Return(header)

    @tornado.gen.coroutine
    def get_body(self):
        """Get the body value from the response.

        :return: a future contains the deserialized value of body
        """

        raw_body = yield get_arg(self, 2)
        if not self.scheme:
            raise tornado.gen.Return(raw_body)
        else:
            body = self.scheme.deserialize_body(raw_body)
            raise tornado.gen.Return(body)

    def set_body_s(self, stream):
        """Set customized body stream.

        Note: the body stream can only be changed before the stream
        is consumed.

        :param stream: InMemStream/PipeStream for body

        :except TChannelError:
            Raise TChannelError if the stream is being sent when you try
            to change the stream.
        """
        if self.argstreams[2].state == StreamState.init:
            self.argstreams[2] = stream
        else:
            raise TChannelError(
                "Unable to change the body since the streaming has started")

    def set_header_s(self, stream):
        """Set customized header stream.

        Note: the header stream can only be changed before the stream
        is consumed.

        :param stream: InMemStream/PipeStream for header

        :except TChannelError:
            Raise TChannelError if the stream is being sent when you try
            to change the stream.
        """

        if self.argstreams[1].state == StreamState.init:
            self.argstreams[1] = stream
        else:
            raise TChannelError(
                "Unable to change the header since the streaming has started")

    def write_header(self, chunk):
        """Write to header.

        Note: the header stream is only available to write before write body.

        :param chunk: content to write to header

        :except TChannelError:
            Raise TChannelError if the response's flush() has been called
        """

        if self.scheme:
            header = self.scheme.serialize_header(chunk)
        else:
            header = chunk

        if self.flushed:
            raise TChannelError("write operation invalid after flush call")

        if (self.argstreams[0].state != StreamState.completed and
                self.argstreams[0].auto_close):
            self.argstreams[0].close()

        return self.argstreams[1].write(header)

    def write_body(self, chunk):
        """Write to header.

        Note: whenever write_body is called, the header stream will be closed.
        write_header method is unavailable.

        :param chunk: content to write to body

        :except TChannelError:
            Raise TChannelError if the response's flush() has been called
        """

        if self.scheme:
            body = self.scheme.serialize_body(chunk)
        else:
            body = chunk

        if self.flushed:
            raise TChannelError("write operation invalid after flush call")

        if (self.argstreams[0].state != StreamState.completed and
                self.argstreams[0].auto_close):
            self.argstreams[0].close()
        if (self.argstreams[1].state != StreamState.completed and
                self.argstreams[1].auto_close):
            self.argstreams[1].close()

        return self.argstreams[2].write(body)

    def flush(self):
        """Flush the response buffer.

        No more write or set operations is allowed after flush call.
        """
        self.flushed = True
        self.close_argstreams()

    def set_exception(self, exception):
        for stream in self.argstreams:
            stream.set_exception(exception)
            stream.close()

    def close_argstreams(self, force=False):
        for stream in self.argstreams:
            if stream.auto_close or force:
                stream.close()



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from thrift.protocol import TBinaryProtocol
from thrift.transport import TTransport

from .. import io
from .. import rw
from .. import scheme


class ThriftArgScheme(scheme.ArgScheme):
    """Represents the ``thrift`` arg scheme.

    It requires a reference to the result type for deserialized objects.
    """

    # Used to serialize and deserialize headers.
    _headers_rw = rw.headers(
        rw.number(2),
        rw.len_prefixed_string(rw.number(2)),
        rw.len_prefixed_string(rw.number(2)),
    )

    def __init__(self, deserialize_type):
        """Initialize a new ThriftArgScheme.

        :param deserialize_type:
            Type of Thrift object contained in the body. This object will be
            deserialized from the stream.
        """
        self.deserialize_type = deserialize_type

    def type(self):
        return 'thrift'

    def serialize_header(self, headers):
        return self._headers_rw.write(headers, io.BytesIO()).getvalue()

    def deserialize_header(self, s):
        return self._headers_rw.read(io.BytesIO(s))

    def serialize_body(self, args):
        trans = TTransport.TMemoryBuffer()
        proto = TBinaryProtocol.TBinaryProtocolAccelerated(trans)
        args.write(proto)

        return trans.getvalue()

    def deserialize_body(self, s):
        trans = TTransport.TMemoryBuffer(s)
        proto = TBinaryProtocol.TBinaryProtocolAccelerated(trans)

        result = self.deserialize_type()
        result.read(proto)
        return result



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from collections import namedtuple

from threadloop import ThreadLoop
from tornado import gen

from tchannel import glossary
from tchannel import tornado as async


class TChannelSyncClient(object):
    """Make synchronous TChannel requests.

    This client does not support incoming connections or requests- this is
    a uni-directional client only.

    The client is implemented on top of the Tornado-based implementation and
    starts and stops IOLoops on-demand.

    .. code-block:: python

        client = TChannelSyncClient()
        response = client.request(
            hostport='localhost:4040',
            service='HelloService',
        ).send(
            'hello', None, json.dumps({"name": "World"})
        )

    """

    def __init__(self, name, process_name=None, known_peers=None, trace=False):
        """Initialize a new TChannelClient.

        :param process_name:
            Name of the calling process. Used for logging purposes only.
        """
        self.async_client = async.TChannel(
            name,
            hostport=glossary.EPHEMERAL_HOSTPORT,
            process_name=process_name,
            known_peers=known_peers,
            trace=trace
        )
        self.threadloop = ThreadLoop()
        self.threadloop.start()

    def request(self, *args, **kwargs):
        """Initiate a new request to a peer.

        :param hostport:
            If specified, requests will be sent to the specific host.
            Otherwise, a known peer will be picked at random.
        :param service:
            Name of the service being called. Defaults to an empty string.
        :param service_threshold:
            If ``hostport`` was not specified, this specifies the score
            threshold at or below which peers will be ignored.
        :returns SyncClientOperation:
            An object with a ``send(arg1, arg2, arg3)`` operation.
        """
        operation = self.async_client.request(*args, **kwargs)
        operation = SyncClientOperation(operation, self.threadloop)

        return operation


class SyncClientOperation(object):
    """Allows making client operation requests synchronously.

    Composes a tchannel.

    This object acts like tchannel.TChannelClientOperation, but instead
    uses a temporary ioloop to make the request synchronously.
    """

    def __init__(self, operation, threadloop):
        assert operation, "operation is required"
        assert threadloop, "threadloop.ThreadLoop is required"
        self.operation = operation
        self.threadloop = threadloop

    def send(self, arg1, arg2, arg3):
        """Send the given triple over the wire.

        :param arg1:
            String containing the contents of arg1. If None, an empty string
            is used.
        :param arg2:
            String containing the contents of arg2. If None, an empty string
            is used.
        :param arg3:
            String containing the contents of arg3. If None, an empty string
            is used.
        :return concurrent.futures.Future:
            Future response from the peer.
        """
        arg1 = arg1 or ''
        arg2 = arg2 or ''
        arg3 = arg3 or ''

        @gen.coroutine
        def make_request():

            response = yield self.operation.send(arg1, arg2, arg3)

            header = yield response.get_header()
            body = yield response.get_body()

            result = Response(header, body)

            raise gen.Return(result)

        future = self.threadloop.submit(make_request)

        return future


Response = namedtuple('Response', 'header, body')



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import tornado

from ..errors import InvalidEndpointError
from ..errors import InvalidMessageError
from ..errors import TChannelError
from ..scheme import RawArgScheme
from .stream import Stream


class ArgSchemeBroker(object):
    """Use serializer to broker request/response."""

    def __init__(self, arg_scheme=None):
        self.endpoint = {}
        self.arg_scheme = arg_scheme or RawArgScheme()

    def register(self, rule, handler):
        """Register handler.

        :param rule: endpoint
        :param handler: endpoint handler
        """
        self.endpoint[rule] = handler

    def handle_call(self, req, resp, proxy):
        if not req.headers.get('as', None) == self.arg_scheme.type():
            raise InvalidMessageError(
                "Invalid arg scheme in request header"
            )

        req.scheme = self.arg_scheme
        resp.scheme = self.arg_scheme

        handler = self.endpoint.get(req.endpoint, None)
        if handler is None:
            raise InvalidEndpointError(
                "Endpoint '%s' for service '%s' is not defined" % (
                    req.endpoint, req.service
                )
            )

        return handler(req, resp, proxy)

    @tornado.gen.coroutine
    def send(self, client, endpoint, header, body, traceflag=None,
             attempt_times=None, ttl=None,
             retry_delay=None):
        """Serialize and deserialize header and body into certain format
            based on arg scheme.
        See parameters definitions in
            :func:`tchannel.tornado.peer.PeerClientOperation.send`
        """
        try:
            if not isinstance(header, Stream):
                raw_header = self.arg_scheme.serialize_header(header)
            else:
                raw_header = header

            if not isinstance(body, Stream):
                raw_body = self.arg_scheme.serialize_body(body)
            else:
                raw_body = body

        except Exception as e:
            raise TChannelError(e.message)

        resp = yield client.send(
            endpoint,
            raw_header,
            raw_body,
            headers={'as': self.arg_scheme.type()},
            traceflag=traceflag,
            attempt_times=attempt_times,
            ttl=ttl,
            retry_delay=retry_delay,
        )

        resp.scheme = self.arg_scheme

        raise tornado.gen.Return(resp)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import inspect
from collections import namedtuple

from thrift import Thrift
from tornado import gen

from tchannel.tornado.broker import ArgSchemeBroker

from .scheme import ThriftArgScheme

# Generated clients will use this base class.
_ClientBase = namedtuple('_ClientBase', 'tchannel hostport service trace')


def client_for(service, service_module, thrift_service_name=None):
    """Build a client class for the given Thrift service.

    The generated class accepts a TChannel and an optional hostport as
    initialization arguments.

    Given ``CommentService`` defined in ``comment.thrift`` and registered with
    Hyperbahn under the name "comment", here's how this may be used:

    .. code-block:: python

        from comment import CommentService

        CommentServiceClient = client_for("comment", CommentService)

        @gen.coroutine
        def post_comment(articleId, msg, hostport=None):
            client = CommentServiceClient(tchannel, hostport)
            yield client.postComment(articleId, CommentService.Comment(msg))

    :param service:
        Name of the Hyperbahn service being called. This is the name with
        which the service registered with Hyperbahn.
    :param service_module:
        The Thrift-generated module for that service. This usually has the
        same name as defined for the service in the IDL.
    :param thrift_service_name:
        If the Thrift service has a different name than its module, use this
        parameter to specify it.
    :returns:
        An object with the same interface as the service that uses the given
        TChannel to call the service.
    """
    assert service_module, 'service_module is required'
    service = service or ''  # may be blank for non-hyperbahn use cases
    if not thrift_service_name:
        thrift_service_name = service_module.__name__.rsplit('.', 1)[-1]

    method_names = [
        name for (name, _) in inspect.getmembers(
            service_module.Iface, predicate=inspect.ismethod
        )
    ]

    def new(cls, tchannel, hostport=None, trace=False):
        """
        :param tchannel:
            TChannel through which the requests will be sent.
        :param hostport:
            Address of the machine to which the requests will be sent, or None
            if the TChannel will do peer selection on a per-request basis.
        """
        return _ClientBase.__new__(cls, tchannel, hostport, service, trace)

    new.__name__ = '__new__'
    methods = {'__new__': new}

    for method_name in method_names:
        methods[method_name] = generate_method(
            service_module, thrift_service_name, method_name
        )

    return type(thrift_service_name + 'Client', (_ClientBase,), methods)


def generate_method(service_module, service_name, method_name):
    """Generate a method for the given Thrift service.

    :param service_module:
        Thrift-generated service module
    :param service_name:
        Name of the Thrift service
    :param method_name:
        Method being called
    """
    assert service_module
    assert service_name
    assert method_name

    args_type = getattr(service_module, method_name + '_args')
    result_type = getattr(service_module, method_name + '_result')
    # TODO result_type is None when the method is oneway.
    # We don't support oneway yet.

    arg_scheme = ThriftArgScheme(result_type)
    result_spec = result_type.thrift_spec
    # result_spec is a tuple of tuples in the form:
    #
    #   (fieldId, fieldType, fieldName, ...)
    #
    # Where "..." is other information we don't care about right now.
    #
    # result_spec will be empty if there is no return value or exception for
    # the method.
    #
    # Its first element, with field ID 0, contains the spec for the return
    # value. It is None if the result type is void but the method may still
    # throw exceptions.
    #
    # Elements after the first one are specs for the exceptions.

    endpoint = '%s::%s' % (service_name, method_name)

    @gen.coroutine
    def send(self, *args, **kwargs):
        params = inspect.getcallargs(
            getattr(service_module.Iface, method_name), self, *args, **kwargs
        )
        params.pop('self')  # self is already known

        # $methodName_args is the implicit struct containing the various
        # method parameters.
        call_args = args_type()
        for name, value in params.items():
            setattr(call_args, name, value)

        response = yield ArgSchemeBroker(arg_scheme).send(
            self.tchannel.request(
                hostport=self.hostport, service=self.service
            ),
            endpoint,
            {},  # TODO: Figure out how to receive headers for the call
            call_args,
            traceflag=self.trace
        )

        call_result = yield response.get_body()
        if not result_spec:
            # void return type and no exceptions allowed
            raise gen.Return(None)

        for exc_spec in result_spec[1:]:
            # May have failed with an exception
            exc = getattr(call_result, exc_spec[2])
            if exc is not None:
                raise exc

        if result_spec[0]:
            # Non-void return type. Return the result.
            success = getattr(call_result, result_spec[0][2])
            if success is not None:
                raise gen.Return(success)
        else:
            # No return type specified and no exceptions raised.
            raise gen.Return(None)

        # Expected a result but nothing was present in the object. Something
        # went wrong.
        raise Thrift.TApplicationException(
            Thrift.TApplicationException.MISSING_RESULT,
            '%s failed: did not receive a result as expected' % method_name
        )
        # TODO: We should probably throw a custom exception instead.

    send.__name__ = method_name
    return send



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from collections import defaultdict

import tornado
import tornado.gen
from tornado import gen

from ..errors import InvalidEndpointError
from ..errors import InvalidMessageError
from ..errors import TChannelError
from ..event import EventType
from ..handler import BaseRequestHandler
from ..messages.error import ErrorCode
from .broker import ArgSchemeBroker
from .response import Response


class RequestDispatcher(BaseRequestHandler):
    """A synchronous RequestHandler that dispatches calls to different
    endpoints based on ``arg1``.

    Endpoints are registered using ``register`` or the ``route``
    decorator.

    .. code-block:: python

        handler = # ...

        @handler.route('my_method')
        def my_method(request, response, proxy):
            response.write('hello world')
    """

    FALLBACK = object()

    def __init__(self):
        super(RequestDispatcher, self).__init__()
        self.default_broker = ArgSchemeBroker()
        self.endpoints = defaultdict(lambda: self.not_found)

    @tornado.gen.coroutine
    def handle_call(self, request, connection):
        # read arg_1 so that handle_call is able to get the endpoint
        # name and find the endpoint handler.
        # the arg_1 value will be store in the request.endpoint field.

        # NOTE: after here, the correct way to access value of arg_1 is through
        # request.endpoint. The original argstream[0] is no longer valid. If
        # user still tries read from it, it will return empty.
        chunk = yield request.argstreams[0].read()

        while chunk:
            request.endpoint += chunk
            chunk = yield request.argstreams[0].read()

        # event: receive_request
        request.tracing.name = request.endpoint

        connection.tchannel.event_emitter.fire(
            EventType.before_receive_request,
            request,
        )

        endpoint = self.endpoints[request.endpoint]

        response = Response(
            id=request.id,
            checksum=request.checksum,
            tracing=request.tracing,
            connection=connection,
            headers={'as': request.headers.get('as', 'raw')},
        )

        connection.post_response(response)

        try:
            yield gen.maybe_future(
                endpoint(
                    request,
                    response,
                    TChannelProxy(
                        connection.tchannel,
                        request.tracing,
                    ),
                )
            )
            response.flush()
        except (InvalidMessageError, InvalidEndpointError) as e:
            response.set_exception(e)
            connection.request_message_factory.remove_buffer(response.id)
            connection.send_error(
                ErrorCode.bad_request,
                e.message,
                response.id,
            )
        except Exception as e:
            response.set_exception(TChannelError(e.message))
            connection.request_message_factory.remove_buffer(response.id)
            connection.send_error(
                ErrorCode.unexpected,
                "An unexpected error has occurred from the handler",
                response.id,
            )

        raise gen.Return(response)

    def route(self, rule, helper=None):
        """See ``register`` for documentation."""

        def decorator(handler):
            self.register(rule, handler, helper)
            return handler

        return decorator

    def register(self, rule, handler, broker=None):
        """Register a new endpoint with the given name.

        .. code-block:: python

            @dispatcher.register('is_healthy')
            def check_health(request, response, proxy):
                # ...

        :param rule:
            Name of the endpoint. Incoming Call Requests must have this as
            ``arg1`` to dispatch to this handler.

            If ``RequestHandler.FALLBACK`` is specified as a rule, the given
            handler will be used as the 'fallback' handler when requests don't
            match any registered rules.

        :param handler:
            A function that gets called with ``Request``, ``Response``, and
            the ``proxy``.

        :param broker:
            Broker injects customized serializer and deserializer into
            request/response object.

            broker==None means it registers as raw handle. It deals with raw
            buffer in the request/response.
        """
        assert handler, "handler must not be None"

        # TODO: Get rid of this Broker thing!! It has the same interface as the
        # Dispatcher but only handles serialization.
        if not broker:
            broker = self.default_broker

        if rule is self.FALLBACK:
            self.endpoints.default_factory = lambda: handler
            return

        broker.register(rule, handler)

        self.endpoints[rule] = broker.handle_call

    @staticmethod
    def not_found(request, response, proxy):
        """Default behavior for requests to unrecognized endpoints."""
        raise InvalidEndpointError(
            "Endpoint '%s' for service '%s' is not defined" % (
                request.endpoint,
                request.service,
            ),
        )


class TChannelProxy(object):
    """TChannel Proxy with additional runtime info

    TChannelProxy contains parent_tracing information which is created by
    received request.

    TChannelProxy will be used as one parameter for the request handler.

    Example::

        def handler(request, response, proxy):

    """
    __slots__ = ('_tchannel', 'parent_tracing')

    def __init__(self, tchannel, parent_tracing=None):
        self._tchannel = tchannel
        self.parent_tracing = parent_tracing

    @property
    def closed(self):
        return self._tchannel.closed

    @property
    def hostport(self):
        return self._tchannel.hostport

    def request(self, hostport=None, service=None, **kwargs):
        kwargs['parent_tracing'] = self.parent_tracing
        return self._tchannel.request(hostport,
                                      service,
                                      **kwargs)
